from typing import Dict
from marshmallow import Schema, fields, validate, validates, ValidationError


class CloudProviderConfigValidator(Schema):
    # TODO: It has to be reworked as soon as we manage multiple cloud provider
    cloud_provider = fields.String(required=True, validate=validate.OneOf(["aws"]))
    region_name = fields.String(required=True)
    aws_access_key_id = fields.String()
    aws_secret_access_key = fields.String()

class CloudQueueWorkerValidator(Schema):
    queue_mapping = fields.Dict(required=True)
    concurrency = fields.Integer(required=True)
    cloud_provider_config = fields.Nested(CloudProviderConfigValidator, required=True)

    @validates("queue_mapping")
    def validate_queue_mapping(self, value: Dict[str, str]):
        if not value:
            raise ValidationError("queue_mapping parameter is missing or empty in the config")



import signal
from time import sleep
from multiprocessing import Process, Event, Queue, Value
from typing import Dict, List, Any, Optional, Tuple, overload
from queue import Empty
from abc import ABC, abstractmethod
import os
import importlib
from uuid import uuid4
from types import FunctionType
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from ..utils import get_client
from ..logging import logger
from .errors import QueueArgumentError, WorkerResultUnknown
from .enums import WorkerResult


class Worker(Process, ABC):
    def __init__(self, internal_queue: Queue, cloud_provider_config: Dict[str, str], *args, **kwargs) -> None:
        """
        Main abstract class used to create read/write workers.
        :param internal_queue: Internal queue used to pass data between workers.
        :param cloud_provider_config: cloud provider config dict used to get appropriate client.
        """
        super().__init__(*args, **kwargs)
        self.internal_queue = internal_queue
        self.cloud_provider_config = cloud_provider_config
        self.exit = Event()

    @abstractmethod
    def run(self):
        pass


class ReadWorker(Worker):
    def __init__(
            self, internal_queue: Queue, cloud_provider_config: Dict[str, str], queue_url: str,
            internal_queue_size: Value
    ) -> None:
        """
        Read worker class. His role is the fetch messages from cloud provider queues
        and send them to internal queue to be processed by executor's workers.
        :param internal_queue: Internal queue used to pass data between workers.
        :param cloud_provider_config: cloud provider config dict used to get appropriate client.
        :param queue_url: Queue's url related to the worker
        :param internal_queue_size: A Value param containing a integer to know the size of the internal queue. We cannot use qsize, Empty or full as it's specified in the multiprocessing documentation as non reliable (https://docs.python.org/3.8/library/multiprocessing.html#multiprocessing.Queue.qsize)
        """
        super().__init__(internal_queue, cloud_provider_config)
        self._queue_url = queue_url
        self.name = queue_url
        self.internal_queue_size = internal_queue_size

    def run(self) -> None:
        """
        Start ReadWorker
        """
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        client = get_client(self.cloud_provider_config)
        logger.debug("The read worker with pid %d", os.getpid())
        while not self.exit.is_set():
            if self.internal_queue_size.value < 50:
                messages = self._fetch_messages(client)
                if messages:
                    self._send_messages_to_executor(messages)
            else:
                sleep(1)
        logger.debug("a signal has been received to close the read worker with pid %d", os.getpid())

    def _fetch_messages(self, client) -> List[Any]:
        """
        Fetch messages from cloud provider queue.
        :param client: Cloud provider client used to fetch messages
        """
        logger.debug("fetch messages from read worker")
        return client.receive_message(QueueUrl=self._queue_url, MaxNumberOfMessages=10, WaitTimeSeconds=20).get(
            "Messages", []
        )

    def _send_messages_to_executor(self, messages: List[Any]) -> None:
        """
        Wrap messages with internal required data and send them to FIFO multiprocessing queue.
        :param messages: A list of messages.
        """
        logger.debug("send messages to executor")
        for message in messages:
            internal_message = {"message": message, "queue_url": self._queue_url}
            self.internal_queue.put(internal_message)
            with self.internal_queue_size.get_lock():
                self.internal_queue_size.value += 1


class ExecutorWorker(Worker):
    def __init__(
            self, queue_mapping: Dict[str, str], internal_queue: Queue, cloud_provider_config: Dict[str, str],
            internal_queue_size: Value
    ) -> None:
        """
        :param queue_mapping: Dict used to make the relation between the cloud provider queue
        and the worker used to process data.
        :param internal_queue: Internal queue used to pass data between workers.
        :param cloud_provider_config: cloud provider config dict used to get appropriate client.
        :param internal_queue_size: A Value param containing a integer to know the size of the internal queue. We cannot use qsize, Empty or full as it's specified in the multiprocessing documentation as non reliable (https://docs.python.org/3.8/library/multiprocessing.html#multiprocessing.Queue.qsize)
        """
        super().__init__(internal_queue, cloud_provider_config)
        self.queue_mapping = queue_mapping
        self.function_mapping = self._initialize_function_mapping()
        self.internal_queue_size = internal_queue_size

    def run(self) -> None:
        """
        Start ExecutorWorker
        """
        logger.debug("The executor worker with pid %d has been started", os.getpid())
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        client = get_client(self.cloud_provider_config)
        while not self.exit.is_set() or not self.internal_queue.empty():
            message = self._fetch_message()
            if message:
                res = self._handle_message(message)
                if not res or res == WorkerResult.ACK:
                    self._acknowledge(message, client)
                elif res[0] == WorkerResult.UNACK:
                    visibility_timeout = res[1].get("visibility_timeout", 30)
                    self._resend_message(message, client, visibility_timeout)
                else:
                    raise WorkerResultUnknown(f"Wrong return type : {type(res[0])}")

        logger.debug("a signal has been received to close the executor worker with pid %d", os.getpid())

    def _fetch_message(self) -> Optional[Dict[str, Any]]:
        """
        Fetch message from internal queue.
        :returns: Dict if there are data to process else None.
        """
        try:
            message = self.internal_queue.get(timeout=0.5)
            with self.internal_queue_size.get_lock():
                if self.internal_queue_size.value > 0:
                    self.internal_queue_size.value -= 1
            return message
        except Empty:
            return None

    def _resend_message(self, message: Dict[str, Any], client, visibility_timeout: int):
        """
        Change the visibility timeout of the message
        :param message: internal queue's message.
        :param client: Cloud provider's client.
        :param visibility_timeout: ew visibility timeout
        :return:
        """
        logger.debug("resend message")
        client.change_message_visibility(
            QueueUrl=message["queue_url"],
            ReceiptHandle=message["message"]["ReceiptHandle"],
            VisibilityTimeout=visibility_timeout
        )

    @overload
    def _handle_message(self, message: Dict[str, Any]) -> Tuple[Literal[WorkerResult.UNACK], Dict[str, Any]]:
        ...

    @overload
    def _handle_message(self, message: Dict[str, Any]) -> Literal[WorkerResult.ACK]:
        ...

    @overload
    def _handle_message(self, message: Dict[str, Any]) -> None:
        ...

    def _handle_message(self, message):
        """
        Execute the function related to the message.
        :param message: internal queue's message.
        :type message: dict
        """
        logger.debug("handle message from executor worker")
        function, kwargs = self.function_mapping[message["queue_url"]]
        return function(message["message"], **kwargs)

    def _acknowledge(self, message: Dict[str, Any], client):
        """
        Acknowledge the message at the end of the function's execution.
        :param message: internal queue's message.
        :param client: Cloud provider's client.
        """
        logger.debug("acknowledge message")
        client.delete_message(QueueUrl=message["queue_url"], ReceiptHandle=message["message"]["ReceiptHandle"])

    def _initialize_function_mapping(self):
        function_mapping = {}
        for key, value in self.queue_mapping.items():
            if isinstance(value, str):
                handler_path = value
                handler_kwargs = {}
            elif isinstance(value, tuple) and len(value) == 2:
                handler_path = value[0]
                handler_kwargs = value[1]
            else:
                raise QueueArgumentError(f"Wrong argument type : {value}")
            function_name = handler_path.split(".")[-1]
            function_path = ".".join(handler_path.split(".")[:-1])

            function_module = importlib.import_module(function_path)
            obj = getattr(function_module, function_name)
            if isinstance(obj, FunctionType):
                function_mapping[key] = (obj, handler_kwargs)
            elif issubclass(obj, AbsCustomWorker):
                instance = obj()
                function_mapping[key] = (instance.run, handler_kwargs)
            else:
                raise QueueArgumentError(f"Wrong object type : {type(obj)}")

        return function_mapping


class AbsCustomWorker(ABC):
    def __init__(self):
        self.worker_id = str(uuid4())

    @abstractmethod
    def run(self, msg, **kwargs):
        pass



from logging import getLogger, StreamHandler, getLevelName, Formatter
from os import getenv


logger = getLogger("cloud_queue_worker")
logger.setLevel(getLevelName(getenv("LOG_LEVEL", "INFO")))
handler = StreamHandler()
formatter = Formatter(
    fmt="%(name)s [%(asctime)s][%(levelname)s]: %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S %p"
)
handler.setFormatter(formatter)
logger.addHandler(handler)



from typing import Optional, Dict, Union
from .validators import CloudQueueWorkerValidator
from .orchestrator import Orchestrator
from .logging import logger


class CloudQueueWorker:
    def __init__(
        self,
        queue_mapping: Optional[Dict[str, str]] = None,
        concurrency: Optional[int] = None,
        cloud_provider_config: Optional[Dict[str, str]] = None,
    ) -> None:
        self.queue_mapping = queue_mapping
        self.concurrency = concurrency
        self.cloud_provider_config = cloud_provider_config

    def update_config(self, config: Dict[str, Union[str, int]]) -> None:
        """
        Update CloudQueueWorker object after initialization using a dict.
        :param config: dict to use to update CloudQueueWorker object
        :type config: dict
        :return: None
        """
        for key, value in config.items():
            setattr(self, key, value)

    def run(self) -> None:
        """
        Method to start CloudQueueWorker
        :return: None
        """
        logger.info("Start cloud queue worker")

        cloud_queue_worker_validator = CloudQueueWorkerValidator().load(self.__dict__)
        orchestrator = Orchestrator(**CloudQueueWorkerValidator().dump(cloud_queue_worker_validator))
        orchestrator.run()



from typing import Dict
import boto3


def get_client(config: Dict[str, str]):
    """
    Get the appropriate client according to configuration sent.
    :param config: The config to get queues from aws, gcp or azure.
    :type config: dict
    :return: The appropriate client to get messages.
    """
    return boto3.client(
        "sqs",
        aws_access_key_id=config.get("aws_access_key_id"),
        aws_secret_access_key=config.get("aws_secret_access_key"),
        region_name=config["region_name"],
        endpoint_url=config.get("endpoint_url", "https://sqs.{}.amazonaws.com".format(config["region_name"])),
    )



from unittest.mock import Mock
from cloud_queue_worker.utils import get_client


def test_get_client(monkeypatch):
    mock_boto_client = Mock()
    mock_config = {
        "aws_access_key_id": "mock_aws_access_key_id",
        "aws_secret_access_key": "mock_aws_secret_access_key",
        "region_name": "mock_region_name",
    }
    monkeypatch.setattr("cloud_queue_worker.utils.boto3.client", mock_boto_client)
    get_client(mock_config)
    mock_boto_client.assert_called_once_with(
        "sqs",
        aws_secret_access_key="mock_aws_secret_access_key",
        aws_access_key_id="mock_aws_access_key_id",
        region_name="mock_region_name",
        endpoint_url="https://sqs.mock_region_name.amazonaws.com"
    )



from unittest.mock import Mock
import pytest
from cloud_queue_worker.orchestrator import Orchestrator
from cloud_queue_worker.errors import ConfigError


def test_check_queue_mapping(monkeypatch):
    monkeypatch.setattr("cloud_queue_worker.orchestrator.ExecutorWorker", Mock())
    monkeypatch.setattr(
        "cloud_queue_worker.orchestrator.get_client",
        Mock(return_value=Mock(list_queues=Mock(return_value={"QueueUrls": ["mock_queue_url"]}))),
    )
    Orchestrator(
        {"mock_queue_url": "mock_method"},
        5,
        {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_access_key_secret": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
    )


def test_check_queue_mapping_with_error(monkeypatch):
    monkeypatch.setattr(
        "cloud_queue_worker.orchestrator.get_client",
        Mock(return_value=Mock(list_queues=Mock(return_value={"QueueUrls": ["invalid_queue"]}))),
    )
    with pytest.raises(ConfigError):
        Orchestrator(
            {"mock_queue_url": "mock_method"},
            5,
            {
                "cloud_provider": "aws",
                "aws_access_key_id": "mock_aws_access_key_id",
                "aws_access_key_secret": "mock_aws_access_key_secret",
                "region_name": "mock_region_name",
            },
        )


def test_initialize_workers(monkeypatch):
    monkeypatch.setattr(
        "cloud_queue_worker.orchestrator.get_client",
        Mock(return_value=Mock(list_queues=Mock(return_value={"QueueUrls": ["mock_queue_url"]}))),
    )
    monkeypatch.setattr("cloud_queue_worker.orchestrator.ExecutorWorker", Mock())
    orchestrator = Orchestrator(
        {"mock_queue_url": "mock_method"},
        5,
        {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_access_key_secret": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
    )
    assert len(orchestrator.read_workers) == 1
    assert len(orchestrator.executor_workers) == 5


def test_run(monkeypatch):
    mock_read_worker_start = Mock()
    mock_read_worker = Mock(return_value=mock_read_worker_start)
    mock_executor_worker_start = Mock()
    mock_executor_worker = Mock(return_value=mock_executor_worker_start)
    monkeypatch.setattr(
        "cloud_queue_worker.orchestrator.get_client",
        Mock(return_value=Mock(list_queues=Mock(return_value={"QueueUrls": ["mock_queue_url"]}))),
    )
    monkeypatch.setattr("cloud_queue_worker.orchestrator.ReadWorker", mock_read_worker)
    monkeypatch.setattr("cloud_queue_worker.orchestrator.ExecutorWorker", mock_executor_worker)
    orchestrator = Orchestrator(
        {"mock_queue_url": "mock_method"},
        5,
        {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_access_key_secret": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
    )
    orchestrator._is_running = False
    orchestrator.run()
    assert mock_read_worker_start.start.call_count == 1
    assert mock_executor_worker_start.start.call_count == 5


def test_stop(monkeypatch):
    mock_read_worker_join = Mock()
    mock_read_worker_exit_set = Mock(set=Mock())
    mock_read_worker_exit = Mock(exit=mock_read_worker_exit_set, join=mock_read_worker_join)
    mock_read_worker = Mock(return_value=mock_read_worker_exit)
    mock_executor_worker_join = Mock()
    mock_executor_worker_exit_set = Mock(set=Mock())
    mock_executor_worker_exit = Mock(exit=mock_executor_worker_exit_set, join=mock_executor_worker_join)
    mock_executor_worker = Mock(return_value=mock_executor_worker_exit)
    mock_close_queue = Mock(close=Mock())
    mock_queue = Mock(return_value=mock_close_queue)
    monkeypatch.setattr(
        "cloud_queue_worker.orchestrator.get_client",
        Mock(return_value=Mock(list_queues=Mock(return_value={"QueueUrls": ["mock_queue_url"]}))),
    )
    monkeypatch.setattr("cloud_queue_worker.orchestrator.ReadWorker", mock_read_worker)
    monkeypatch.setattr("cloud_queue_worker.orchestrator.ExecutorWorker", mock_executor_worker)
    monkeypatch.setattr("cloud_queue_worker.orchestrator.Queue", mock_queue)
    orchestrator = Orchestrator(
        {"mock_queue_url": "mock_method"},
        5,
        {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_access_key_secret": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
    )
    orchestrator._stop()
    assert mock_read_worker_exit_set.set.call_count == 1
    assert mock_executor_worker_exit_set.set.call_count == 5
    assert mock_executor_worker_join.call_count == 5
    assert mock_read_worker_join.call_count == 1
    mock_close_queue.close.assert_called_once_with()



from unittest.mock import Mock
import pytest
from marshmallow.exceptions import ValidationError
from cloud_queue_worker.queue_worker import CloudQueueWorker


def test_update_config():
    new_config = {
        "queue_mapping": {"mock_queue_url": "function_to_call"},
        "concurrency": 3,
        "cloud_provider_config": {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_access_key_secret": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
    }
    cloud_queue_worker = CloudQueueWorker()
    cloud_queue_worker.update_config(new_config)
    assert vars(cloud_queue_worker) == new_config


@pytest.mark.parametrize(
    "config",
    [
        {
            "queue_mapping": {"mock_queue_url": "function_to_call"},
            "cloud_provider_config": {
                "cloud_provider": "aws",
                "aws_access_key_id": "mock_aws_access_key_id",
                "aws_access_key_secret": "mock_aws_access_key_secret",
                "region_name": "mock_region_name",
            },
        },
        {
            "concurrency": 2,
            "cloud_provider_config": {
                "cloud_provider": "aws",
                "aws_access_key_id": "mock_aws_access_key_id",
                "aws_access_key_secret": "mock_aws_access_key_secret",
                "region_name": "mock_region_name",
            },
        },
        {
            "queue_mapping": {"mock_queue_url": "function_to_call"},
            "concurrency": 2,
        },
        {
            "queue_mapping": {"mock_queue_url": "function_to_call"},
            "cloud_provider_config": {
                "cloud_provider": "invalid_cloud_provider",
                "aws_access_key_id": "mock_aws_access_key_id",
                "aws_access_key_secret": "mock_aws_access_key_secret",
                "region_name": "mock_region_name",
            },
        },
        {
            "queue_mapping": {"mock_queue_url": "function_to_call"},
            "cloud_provider_config": {
                "cloud_provider": "aws",
                "aws_access_key_id": "mock_aws_access_key_id",
                "aws_access_key_secret": "mock_aws_access_key_secret",
            },
        },
    ],
)
def test_run_with_invalid_config(config):
    cloud_queue_worker = CloudQueueWorker()
    cloud_queue_worker.update_config(config)
    with pytest.raises(ValidationError):
        cloud_queue_worker.run()


def test_run(monkeypatch):
    mock_run = Mock()
    mock_orchestrator = Mock(return_value=mock_run)
    new_config = {
        "queue_mapping": {"mock_queue_url": "function_to_call"},
        "concurrency": 3,
        "cloud_provider_config": {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_secret_access_key": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
    }
    cloud_queue_worker = CloudQueueWorker()
    cloud_queue_worker.update_config(new_config)
    monkeypatch.setattr("cloud_queue_worker.queue_worker.Orchestrator", mock_orchestrator)
    cloud_queue_worker.run()
    mock_run.run.assert_called_once_with()



from multiprocessing import Queue, Value
from unittest.mock import Mock, MagicMock
import pytest
from cloud_queue_worker.worker.worker import Worker, ReadWorker, ExecutorWorker, AbsCustomWorker


def mock_function(message, **kwargs):
    pass


def test_worker():
    with pytest.raises(TypeError):
        Worker(
            Queue(),
            {
                "cloud_provider": "aws",
                "aws_access_key_id": "mock_aws_access_key_id",
                "aws_access_key_secret": "mock_aws_access_key_secret",
                "region_name": "mock_region_name",
            },
        )


def test_read_worker_fetch_messages():
    worker = ReadWorker(
        Queue(),
        {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_access_key_secret": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
        "mock_queue_url",
        Value("i", 0),
    )
    mock_messages = Mock(return_value={"Messages": ["test"]})
    mock_client = Mock(receive_message=mock_messages)
    result = worker._fetch_messages(mock_client)
    mock_client.receive_message.assert_called_once_with(
        QueueUrl="mock_queue_url",
        MaxNumberOfMessages=10,
        WaitTimeSeconds=20,
    )
    assert result == ["test"]


def test_send_messages_to_executor():
    mock_put_in_queue = Mock()
    mock_queue = Mock(put=mock_put_in_queue)
    worker = ReadWorker(
        mock_queue,
        {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_access_key_secret": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
        "mock_queue_url",
        Value("i", 0),
    )
    worker._send_messages_to_executor(["message1"])
    mock_put_in_queue.assert_called_once_with({"message": "message1", "queue_url": "mock_queue_url"})
    assert worker.internal_queue_size.value == 1


def test_executor_worker_fetch_message(monkeypatch):
    mock = Mock(import_module=Mock(return_value=Mock(mock_method=mock_function)))
    monkeypatch.setattr("cloud_queue_worker.worker.worker.importlib", mock)
    executor_queue = Queue()
    try:
        executor_queue.put("message")
        worker = ExecutorWorker(
            {"mock_queue_url": "mock_method"},
            executor_queue,
            {
                "cloud_provider": "aws",
                "aws_access_key_id": "mock_aws_access_key_id",
                "aws_access_key_secret": "mock_aws_access_key_secret",
                "region_name": "mock_region_name",
            },
            Value("i", 1),
        )
        result = worker._fetch_message()
        assert result == "message"
        assert worker.internal_queue_size.value == 0
    finally:
        executor_queue.close()


def test_executor_worker_fetch_message_without_message_in_internal_queue(monkeypatch):
    mock = Mock(import_module=Mock(return_value=Mock(mock_method=mock_function)))
    monkeypatch.setattr("cloud_queue_worker.worker.worker.importlib", mock)
    executor_queue = Queue()
    try:
        worker = ExecutorWorker(
            {"mock_queue_url": "mock_method"},
            executor_queue,
            {
                "cloud_provider": "aws",
                "aws_access_key_id": "mock_aws_access_key_id",
                "aws_access_key_secret": "mock_aws_access_key_secret",
                "region_name": "mock_region_name",
            },
            Value("i", 0),
        )
        result = worker._fetch_message()
        assert result is None
    finally:
        executor_queue.close()


def test_acknowledge(monkeypatch):
    mock = Mock(import_module=Mock(return_value=Mock(mock_method=mock_function)))
    monkeypatch.setattr("cloud_queue_worker.worker.worker.importlib", mock)
    mock_client = Mock()
    worker = ExecutorWorker(
        {"mock_queue_url": "mock_method"},
        Queue(),
        {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_access_key_secret": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
        Value("i", 1),
    )
    worker._acknowledge({"queue_url": "mock_queue_url", "message": {"ReceiptHandle": "mockReceiptHandle"}}, mock_client)
    mock_client.delete_message.assert_called_once_with(QueueUrl="mock_queue_url", ReceiptHandle="mockReceiptHandle")


def test_resend_message(monkeypatch):
    mock = Mock(import_module=Mock(return_value=Mock(mock_method=mock_function)))
    monkeypatch.setattr("cloud_queue_worker.worker.worker.importlib", mock)
    mock_client = Mock()
    worker = ExecutorWorker(
        {"mock_queue_url": "mock_method"},
        Queue(),
        {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_access_key_secret": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
        Value("i", 1),
    )
    worker._resend_message({"queue_url": "mock_queue_url", "message": {"ReceiptHandle": "mockReceiptHandle"}},
                           mock_client, 15)
    mock_client.change_message_visibility.assert_called_once_with(QueueUrl="mock_queue_url",
                                                                  ReceiptHandle="mockReceiptHandle",
                                                                  VisibilityTimeout=15)


def test_handle_message(monkeypatch):
    mock_function = Mock()

    def mock_function2(msg, **kwargs):
        mock_function(msg, **kwargs)

    mock = Mock(import_module=Mock(return_value=Mock(mock_method=mock_function2)))
    monkeypatch.setattr("cloud_queue_worker.worker.worker.importlib", mock)
    worker = ExecutorWorker(
        {"mock_queue_url": "mymockmodule.mock_method"},
        Queue(),
        {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_access_key_secret": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
        Value("i", 0),
    )
    worker._handle_message({"queue_url": "mock_queue_url", "message": {"ReceiptHandle": "mockReceiptHandle"}})
    mock_function.assert_called_once_with({"ReceiptHandle": "mockReceiptHandle"})


def test_handle_message_with_args(monkeypatch):
    mock_function = Mock()

    def mock_function2(msg, **kwargs):
        mock_function(msg, **kwargs)

    mock = Mock(import_module=Mock(return_value=Mock(mock_method=mock_function2)))
    monkeypatch.setattr("cloud_queue_worker.worker.worker.importlib", mock)
    worker = ExecutorWorker(
        {"mock_queue_url": ("mymockmodule.mock_method", {"arg_name": "arg_value"})},
        Queue(),
        {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_access_key_secret": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
        Value("i", 0),
    )
    worker._handle_message({"queue_url": "mock_queue_url", "message": {"ReceiptHandle": "mockReceiptHandle"}})
    mock_function.assert_called_once_with({"ReceiptHandle": "mockReceiptHandle"}, arg_name="arg_value")


def test_execute_worker_with_class(monkeypatch):
    mock_run = Mock()

    class MockCustomWorker(AbsCustomWorker):
        def run(self, msg, **kwargs):
            mock_run(msg, **kwargs)

    mock = Mock(import_module=Mock(return_value=Mock(mock_method=MockCustomWorker)))

    monkeypatch.setattr("cloud_queue_worker.worker.worker.importlib", mock)
    worker = ExecutorWorker(
        {"mock_queue_url": "mymockmodule.mock_method"},
        Queue(),
        {
            "cloud_provider": "aws",
            "aws_access_key_id": "mock_aws_access_key_id",
            "aws_access_key_secret": "mock_aws_access_key_secret",
            "region_name": "mock_region_name",
        },
        Value("i", 0),
    )
    worker._handle_message({"queue_url": "mock_queue_url", "message": {"ReceiptHandle": "mockReceiptHandle"}})
    mock_run.assert_called_once_with({"ReceiptHandle": "mockReceiptHandle"})



from setuptools import setup, find_packages

setup(
    name="cloud_queue_worker",
    version="0.0.10",
    author="Enlaps Open Source",
    author_email="contact@enlaps.fr",
    description="Library to create workers for aws, azure and gcp queue services",
    url="https://gitlab.com/enlaps-public/web/cloud_queue_worker",
    install_requires=["boto3>=1.14.63", "marshmallow>=3.8.0"],
    packages=find_packages(),
    extras_require={"dev": ["pytest", "mypy", "pylint"]},
    python_requires=">=3.6",
)



import time
from datetime import datetime, timedelta
import signal
from multiprocessing import Queue, Value
from typing import Dict, List
from .worker.worker import ReadWorker, ExecutorWorker
from .errors import ConfigError
from .utils import get_client
from .logging import logger


class Orchestrator:
    """
    The orchestrator role is to create read and executor workers and an internal queue to send data between them.
    """

    def __init__(self, queue_mapping: Dict[str, str], concurrency: int, cloud_provider_config: Dict[str, str]) -> None:
        self.cloud_provider_config = cloud_provider_config
        self._check_queue_mapping(queue_mapping)
        self.queue_mapping = queue_mapping
        self.concurrency = concurrency
        self.internal_queue: Queue = Queue()
        self.internal_queue_size = Value("i", 0)
        self.read_workers: List[ReadWorker] = []
        self.executor_workers: List[ExecutorWorker] = []
        self._is_running = True
        self._initialize_workers()
        self._register_shutdown_signals()

    def _check_queue_mapping(self, queue_mapping: Dict[str, str]) -> None:
        """
        Check if queue urls specified in queue_mapping dict by user exist for the cloud provider.
        :param queue_mapping: Dict which contain relation between queue urls and methods to execute.
        :type queue_mapping: dict
        """
        client = get_client(self.cloud_provider_config)
        queue_urls = client.list_queues().get("QueueUrls", [])
        for key in queue_mapping:
            if key not in queue_urls:
                raise ConfigError("The queue url {} does not exist".format(key))

    def _initialize_workers(self) -> None:
        """
        Create read and executor workers.
        """
        for queue_url in self.queue_mapping:
            self.read_workers.append(
                ReadWorker(
                    internal_queue=self.internal_queue,
                    internal_queue_size=self.internal_queue_size,
                    queue_url=queue_url,
                    cloud_provider_config=self.cloud_provider_config,
                )
            )
        for _ in range(self.concurrency):
            self.executor_workers.append(
                ExecutorWorker(
                    queue_mapping=self.queue_mapping,
                    internal_queue=self.internal_queue,
                    internal_queue_size=self.internal_queue_size,
                    cloud_provider_config=self.cloud_provider_config,
                )
            )

    def run(self) -> None:
        """
        Run orchestrator
        """
        for read_worker in self.read_workers:
            read_worker.start()

        for executor_worker in self.executor_workers:
            executor_worker.start()

        self._check_workers_statuses()

    def _stop(self) -> None:
        """
        Stop read and executor workers. The orchestrator try to clean up his internal queue before stop the workers.
        If the queue is not cleaned after 30 seconds, data remaining is lost.
        """
        for read_worker in self.read_workers:
            read_worker.exit.set()

        for executor_worker in self.executor_workers:
            executor_worker.exit.set()

        queue_vacuum_start = datetime.now()
        while datetime.now() - queue_vacuum_start <= timedelta(seconds=30):
            if self.internal_queue.empty():
                break
            time.sleep(1)

        self.internal_queue.close()

        for read_worker in self.read_workers:
            read_worker.join()

        for executor_worker in self.executor_workers:
            executor_worker.join()

    def _check_workers_statuses(self) -> None:
        """
        Check if workers are still alive. If not, respawn them.
        """
        while self._is_running:
            time.sleep(1)
            read_workers_to_remove = [read_worker for read_worker in self.read_workers if not read_worker.is_alive()]
            executor_workers_to_remove = [
                executor_worker for executor_worker in self.executor_workers if not executor_worker.is_alive()
            ]

            for read_worker_to_remove in read_workers_to_remove:
                self.read_workers.remove(read_worker_to_remove)
                new_read_worker = ReadWorker(
                    self.internal_queue,
                    self.cloud_provider_config,
                    read_worker_to_remove.name,
                    self.internal_queue_size,
                )
                self.read_workers.append(new_read_worker)
                new_read_worker.start()

            for executor_worker_to_remove in executor_workers_to_remove:
                self.executor_workers.remove(executor_worker_to_remove)
                new_executor_worker = ExecutorWorker(
                    self.queue_mapping, self.internal_queue, self.cloud_provider_config, self.internal_queue_size
                )
                self.executor_workers.append(new_executor_worker)
                new_executor_worker.start()

        self._stop()

    def _register_shutdown_signals(self) -> None:
        """
        Register multiple shutdown signals
        """
        for sig in [signal.SIGINT, signal.SIGTERM, signal.SIGQUIT, signal.SIGHUP]:
            signal.signal(sig, self._stop_orchestrator)

    def _stop_orchestrator(self, _signum, _frame) -> None:
        """
        Set _is_running variable to false.
        """
        logger.info("a signal has been received. Stop orchestrator...")
        self._is_running = False
